@@ -4,38 +4,134 @@ import pathlib
 
 import numpy
 
+from . import _compatibility
 from .box import Box
 from .data import _readline
 from .snapshot import Snapshot
 
+if _compatibility.pyzstd_version is not None:
+    import pyzstd
+
 
 class DumpFile:
     """LAMMPS dump file.
 
-    The dump file is a flexible file format, so a ``schema`` can be given
-    to parse the atom data. The ``schema`` is given as a dictionary of column
-    indexes. Valid keys for the schema match the names and shapes in the `Snapshot`.
-    The keys requiring only 1 column index are: ``id``, ``typeid``, ``molecule``,
-    ``charge``, and ``mass``. The keys requiring 3 column indexes are ``position``,
-    ``velocity``, and ``image``. If a ``schema`` is not specified, it will be deduced
-    from the ``ITEM: ATOMS`` header.
+    The LAMMPS dump file is a highly flexible file format. The ``schema`` of the
+    file is deduced from the ``ITEM: ATOMS`` header unless one is manually
+    specified. If specificed, the ``schema`` must be a dictionary mapping pieces
+    of data to column indexes (0-indexed). Valid keys for the schema match the
+    names and shapes in the `Snapshot`. The keys requiring only 1 column index
+    are:
 
-    The vector-valued fields (``position``, ``velocity``, ``image``) must contain all
-    three elements.
+    - ``"id"``
+    - ``"typeid"``
+    - ``"molecule"``
+    - ``"charge"``
+    - ``"mass"``
+
+    The keys requiring 3 column indexes are:
+
+    - ``"position"``
+    - ``"velocity"``
+    - ``"image"``
+
+    LAMMPS dumps particles in an unknown order unless you have used the
+    ``dump_modify sort`` option. If you want particles to be ordered by ``id``
+    in the `Snapshot`, use ``sort_ids=True`` (default). Note that slightly
+    faster reading may be possible by setting this option to ``False``.
+
+    A `DumpFile` is iterable, so you can use it to go through all the snapshots
+    of a trajectory. It supports LAMMPS dump file with different compression
+    formats such as gzip and Zstandard
+    (`pyzstd <https://pyzstd.readthedocs.io/en/stable/>`_ package needs to be installed
+    separately for Zstandard support).
+    Random access to snapshots is not currently implemented, but it may be added
+    in future. If you want to randomly access snapshots, you should load the whole
+    file into a list, but the memory requirements to do so may be large.
+
+    The dump file may not contain certain information about your particles, for
+    example, topology, or you may choose not to write this information in the
+    dump file because it does not change frame-to-frame. The `copy_from` option
+    allows this information to be copied into a new snapshot from a reference
+    one, e.g, that was read from a `DataFile`.
 
     Parameters
     ----------
     filename : str
         Path to dump file.
     schema : dict
-        Schema for the contents of the file. Defaults to ``None``, which means to read
-        it from the file.
+        Schema for the contents of the file. Defaults to ``None``, which means
+        to read it from the file.
     sort_ids : bool
-        If true, sort the particles by ID in each snapshot.
-    copy_from : :class:`Snapshot`
-        If specified, copy fields that are missing in the dump file but are set in
-        a reference :class:`Snapshot`. The fields that can be copied are ``typeid``,
-        ``molecule``, ``charge``, and ``mass``.
+        If ``True``, sort the particles by ID in each snapshot.
+    copy_from : `Snapshot`
+        If specified, copy supported fields that are missing in the dump file
+        but are set in a reference `Snapshot`.
+
+    Example
+    -------
+
+    .. skip: start(lammps == None, reason="lammps not installed")
+
+    .. invisible-code-block: python
+
+        filename = tmp_path / "atoms.lammpstrj"
+        lmps = lammps.lammps(cmdargs=["-log", "none", "-nocite"])
+
+        cmds = []
+        cmds += ["units lj"]
+        cmds += ["atom_style atomic"]
+        cmds += ["boundary p p p"]
+
+        box_length = 3
+
+        cmds += [f"region box block 0 {box_length} 0 {box_length} 0 {box_length}"]
+        cmds += ["create_box 1 box"]
+
+        cmds += ["create_atoms 1 random 10 12345 box"]
+        cmds += ["mass 1 1.0"]
+
+        cmds += ["pair_style lj/cut 2.5"]
+        cmds += ["pair_coeff 1 1 1.0 1.0 2.5"]
+
+        cmds += ["minimize 1.0e-4 1.0e-6 100 1000"]
+        cmds += ["timestep 0.005"]
+        cmds += ["fix 1 all nvt temp 1.0 1.0 0.1"]
+        cmds += ["thermo 1000"]
+        cmds += ["thermo_style custom step temp pe ke etotal press density"]
+        cmds += [f"dump 1 all custom 100 {filename} id x y z ix iy iz"]
+        cmds += ["run 1000"]
+
+        lmps.commands_list(cmds)
+        lmps.close()
+
+    Create a dump file object:
+
+    .. code-block:: python
+
+        traj = lammpsio.DumpFile(filename)
+
+    Iterate snapshots:
+
+    .. code-block:: python
+
+        for snapshot in traj:
+            print(snapshot.step)
+
+    You can also get the number of snapshots in the `DumpFile`, but this does
+    require reading the entire file: use with caution!
+
+    .. code-block:: python
+
+        num_frames = len(traj)
+
+    Random access by creating a list:
+
+
+    .. code-block:: python
+
+        snapshots = [snap for snap in traj]
+        print(snapshots[3].step)
 
     """
 
@@ -56,14 +152,26 @@ class DumpFile:
             Path to dump file.
         schema : dict
             Schema for the contents of the file.
-        snapshots : :class:`Snapshot` or list
+        snapshots : `Snapshot` or list
             One or more snapshots to write to the dump file.
 
         Returns
         -------
-        :class:`DumpFile`
+        `DumpFile`
             The object representing the new dump file.
 
+        Example
+        -------
+        A `DumpFile` can be created from a list of snapshots:
+
+        .. code-block:: python
+
+            schema = {"id":0, "position":(1, 2, 3), "image": (4, 5, 6)}
+
+            lammpsio.DumpFile.create(filename, schema, snapshots)
+
+        .. skip: end
+
         """
         # map out the schema into a dump row
         # each entry is a tuple: (column, (attribute, index))
@@ -92,10 +200,9 @@ class DumpFile:
                 f.write(f"{snap.N}\n")
 
                 # always assume periodic in all directions
-                box_header = "ITEM: BOX BOUNDS pp pp pp"
                 if snap.box.tilt is not None:
+                    f.write("ITEM: BOX BOUNDS xy xz yz pp pp pp\n")
                     xy, xz, yz = snap.box.tilt
-                    box_header += f" {xy:f} {xz:f} {yz:f}"
                     lo = [
                         snap.box.low[0] + min([0.0, xy, xz, xy + xz]),
                         snap.box.low[1] + min([0.0, yz]),
@@ -106,12 +213,14 @@ class DumpFile:
                         snap.box.high[1] + max([0.0, yz]),
                         snap.box.high[2],
                     ]
+                    for i in range(3):
+                        f.write(f"{lo[i]:f} {hi[i]:f} {snap.box.tilt[i]:f}\n")
                 else:
+                    f.write("ITEM: BOX BOUNDS pp pp pp\n")
                     lo = snap.box.low
                     hi = snap.box.high
-                f.write(box_header + "\n")
-                for i in range(3):
-                    f.write(f"{lo[i]:f} {hi[i]:f}\n")
+                    for i in range(3):
+                        f.write(f"{lo[i]:f} {hi[i]:f}\n")
 
                 # mapping from lammpsio to LAMMPS dump keys
                 lammps_fields = {
@@ -157,16 +266,42 @@ class DumpFile:
                     f.write(line + "\n")
 
         filename_path = pathlib.Path(filename)
-        if filename_path.suffix == ".gz":
+        compression = cls._compression_from_suffix(filename_path.suffix)
+        if compression:
             tmp = pathlib.Path(filename).with_suffix(filename_path.suffix + ".tmp")
-            with open(filename, "rb") as src, gzip.open(tmp, "wb") as dest:
+            with open(filename, "rb") as src, compression.open(tmp, "wb") as dest:
                 dest.writelines(src)
             os.replace(tmp, filename)
 
         return DumpFile(filename, schema)
 
+    @staticmethod
+    def _compression_from_suffix(suffix):
+        if suffix == ".gz":
+            return gzip
+        elif suffix == ".zst":
+            if _compatibility.pyzstd_version is None:
+                raise ModuleNotFoundError("pyzstd needed for zstd compression")
+            return pyzstd
+        else:
+            return None
+
     @property
     def copy_from(self):
+        """`Snapshot`: Copy fields that are missing in a dump file.
+
+        The fields that can be copied are:
+
+        - `Snapshot.angles`
+        - `Snapshot.bonds`
+        - `Snapshot.charge`
+        - `Snapshot.dihedrals`
+        - `Snapshot.impropers`
+        - `Snapshot.mass`
+        - `Snapshot.molecule`
+        - `Snapshot.typeid`
+
+        """
         return self._copy_from
 
     @copy_from.setter
@@ -183,7 +318,7 @@ class DumpFile:
     @filename.setter
     def filename(self, value):
         self._filename = value
-        self._gzip = pathlib.Path(value).suffix == ".gz"
+        self._compression = self._compression_from_suffix(pathlib.Path(value).suffix)
 
         # configure section labels of dump file
         self._section = {
@@ -192,7 +327,7 @@ class DumpFile:
             "box": "ITEM: BOX BOUNDS",
             "atoms": "ITEM: ATOMS",
         }
-        if self._gzip:
+        if self._compression:
             for key, val in self._section.items():
                 self._section[key] = val.encode()
 
@@ -215,8 +350,8 @@ class DumpFile:
 
     def _open(self):
         """Open the file handle for reading."""
-        if self._gzip:
-            f = gzip.open(self.filename, "rb")
+        if self._compression:
+            f = self._compression.open(self.filename, "rb")
         else:
             f = open(self.filename, "r")
         return f
@@ -259,19 +394,20 @@ class DumpFile:
                     box_header = line.split()
                     # check for triclinic
                     if len(box_header) == 9:
-                        box_tilt = [float(x) for x in box_header[6:9]]
+                        is_triclinic = True
                     elif len(box_header) == 6:
-                        box_tilt = None
+                        is_triclinic = False
                     else:
                         raise IOError("Incorrectly formed box bound header")
-                    box_x = _readline(f, True)
-                    box_y = _readline(f, True)
-                    box_z = _readline(f, True)
-                    x_lo, x_hi = [float(x) for x in box_x.split()]
-                    y_lo, y_hi = [float(y) for y in box_y.split()]
-                    z_lo, z_hi = [float(z) for z in box_z.split()]
-                    if box_tilt is not None:
-                        xy, xz, yz = box_tilt
+                    box_ = [
+                        [float(v) for v in _readline(f, True).split()]
+                        for line_ in range(3)
+                    ]
+                    x_lo, x_hi = box_[0][:2]
+                    y_lo, y_hi = box_[1][:2]
+                    z_lo, z_hi = box_[2][:2]
+                    if is_triclinic:
+                        xy, xz, yz = [row[2] for row in box_]
                         lo = [
                             x_lo - min([0.0, xy, xz, xy + xz]),
                             y_lo - min([0.0, yz]),
@@ -282,7 +418,7 @@ class DumpFile:
                             y_hi - max([0.0, yz]),
                             z_hi,
                         ]
-                        box = Box(lo, hi, box_tilt)
+                        box = Box(lo, hi, [xy, xz, yz])
                     else:
                         box = Box([x_lo, y_lo, z_lo], [x_hi, y_hi, z_hi])
 
@@ -290,7 +426,7 @@ class DumpFile:
                 if state == 3 and self._section["atoms"] in line:
                     state += 1
 
-                    # extract the scehma
+                    # extract the schema
                     if self.schema is None:
                         # mapping from LAMMPS dump keys to lammpsio
                         lammps_fields = {
@@ -312,7 +448,7 @@ class DumpFile:
                         schema = {}
                         schema_header = line.split()[2:]
                         for i, field in enumerate(schema_header):
-                            if self._gzip:
+                            if self._compression:
                                 field = field.decode()
                             if field in lammps_fields:
                                 key, key_idx = lammps_fields[field]
@@ -323,7 +459,7 @@ class DumpFile:
                                 else:
                                     schema[key] = i
                         # validate tuple
-                        for key in ("position", "velocity", " image"):
+                        for key in ("position", "velocity", "image"):
                             if key in schema and any(x is None for x in schema[key]):
                                 raise IOError("lammpsio requires 3-element vectors")
                         self.schema = schema
@@ -389,6 +525,14 @@ class DumpFile:
                             snap.charge = self._copy_from.charge[copy_id]
                         if not snap.has_mass() and self._copy_from.has_mass():
                             snap.mass = self._copy_from.mass[copy_id]
+                        if self._copy_from.has_bonds():
+                            snap.bonds = self._copy_from.bonds
+                        if self._copy_from.has_angles():
+                            snap.angles = self._copy_from.angles
+                        if self._copy_from.has_dihedrals():
+                            snap.dihedrals = self._copy_from.dihedrals
+                        if self._copy_from.has_impropers():
+                            snap.impropers = self._copy_from.impropers
 
                     yield snap
                     del snap, N, box, step
